unit Kr_Graphic;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, dsgnIntf;

const
  CONST_MAXELEMENTS=50;
  CONST_MAXPOINTS=2000;

type
  KrAbout=class;
  KrGBorder=(boLine, boUp, boDown, boNone);
  KrGLines=(liHor, liVert, liNone, liVertHor);
  KrGStatus=(stBorderMin, stBorderMax, stAlarmMin, stAlarmMax, stWarningMin, stWarningMax, stNone, stNormal);

  KrAlarmMin=procedure(Sender:TObject; Element:integer; Value:int64) of object;
  KrAlarmMax=procedure(Sender:TObject; Element:integer; Value:int64) of object;
  KrWarningMin=procedure(Sender:TObject; Element:integer; Value:int64) of object;
  KrWarningMax=procedure(Sender:TObject; Element:integer; Value:int64) of object;
  KrBorderMin=procedure(Sender:TObject; Element:integer; Value:int64) of object;
  KrBorderMax=procedure(Sender:TObject; Element:integer; Value:int64) of object;
  KrAfterPoint=procedure(Sender:TObject; Element:integer; Value:int64) of object;
  KrBeforePoint=procedure(Sender:TObject; Element:integer; Value:int64) of object;

  KrElements=Record
    Visible:boolean;
    AlarmMin:int64;
    AlarmMax:int64;
    WarningMin:int64;
    WarningMax:int64;
    BorderMin:int64;
    BorderMax:int64;
    Color:TColor;
    Status:KrGStatus;
    Points:array of int64;
  end;
  pKrElements=^KrElements;

  KrGraphic = class(TCustomControl)
  private
    _About:KrAbout;
    _Color:TColor;
    _Border:KrGBorder;
    _Interval:integer;
    _LineColor:TColor;
    _ShadowColor:TColor;
    _LightColor:TColor;
    _LinesHor:integer;
    _LinesVert:integer;
    _Lines:KrGLines;
    _LinesVertHorColor:TColor;
    _TimerEnabled:boolean;

    _AlarmMin:KrAlarmMin;
    _AlarmMax:KrAlarmMax;
    _WarningMin:KrWarningMin;
    _WarningMax:KrWarningMax;
    _BorderMin:KrBorderMin;
    _BorderMax:KrBorderMax;
    _AfterPoint:KrAfterPoint;
    _BeforePoint:KrBeforePoint;

    _OnTmrTimer:TNotifyEvent;

    Elements:array[1..CONST_MAXELEMENTS] of pKrElements;
    Tmr:TTimer;
    BM:TBitmap;

    procedure BeginPaint;
    procedure TmrTimer(Sender:TObject);
  protected
    procedure SetColor(value:TColor);
    procedure SetBorder(value:KrGBorder);
    procedure SetInterval(value:integer);
    procedure SetLineColor(value:TColor);
    procedure SetShadowColor(value:TColor);
    procedure SetLightColor(value:TColor);
    procedure SetLinesHor(value:integer);
    procedure SetLinesVert(value:integer);
    procedure SetLines(value:KrGLines);
    procedure SetLinesVertHorColor(value:TColor);
    procedure SetTimerEnabled(value:boolean);

    procedure Paint; override;

    procedure PaintBorder;
    procedure PaintLines;
    procedure PaintAlarmLines;
    procedure PaintPoints;

    function CalcPointX(Element:integer; i:int64):integer;
  public
    CONST_NOTDEFINED:int64;

    constructor Create(AOwner:TComponent); override;
    destructor  Destroy; override;

    procedure ResetPoints(Element:integer);
    procedure DeleteElement(Element:integer);
    procedure SetAlarmMin(Element:integer; Value:int64);
    procedure SetAlarmMax(Element:integer; Value:int64);
    procedure SetBorderMin(Element:integer; Value:int64);
    procedure SetBorderMax(Element:integer; Value:int64);
    procedure SetWarningMin(Element:integer; Value:int64);
    procedure SetWarningMax(Element:integer; Value:int64);
    procedure SetColorLine(Element:integer; Value:TColor);
    procedure PointValue(Element, Point:integer; Value:int64); overload;
    procedure IsVisible(Element:integer; Value:boolean); overload;
    procedure SendAlarms(Element:integer; Value:int64);

    function GetAlarmMin(Element:integer):int64;
    function GetAlarmMax(Element:integer):int64;
    function GetWarningMin(Element:integer):int64;
    function GetWarningMax(Element:integer):int64;
    function GetBorderMin(Element:integer):int64;
    function GetBorderMax(Element:integer):int64;
    function GetColorLine(Element:integer):TColor;
    function GetStatus(Element:integer):KrGStatus;
    function PointValue(Element, Point:integer):int64; overload;
    function CreateElement(No:integer; BorderMin, BorderMax, AlarmMin, AlarmMax, WarningMin, WarningMax:int64; Color:TColor):boolean;
    function IsDefined(Element, Point:integer):boolean;
    function IsVisible(Element:integer):boolean; overload;
    function AppendPointValue(Element:integer; value:int64):int64;
  published
    property About:KrAbout read _About write _About;
    property BgColor:TColor read _Color write SetColor default clBlack;
    property Border:KrGBorder read _Border write SetBorder default boLine;
    property Interval:integer read _Interval write SetInterval default 1000;
    property LineColor:TColor read _LineColor write SetLineColor default clWhite;
    property ShadowColor:TColor read _ShadowColor write SetShadowColor default clGray;
    property LightColor:TColor read _LightColor write SetLightColor default clWhite;
    property LinesHor:integer read _LinesHor write SetLinesHor default 10;
    property LinesVert:integer read _LinesVert write SetLinesVert default 10;
    property Lines:KrGLines read _Lines write SetLines default liVertHor;
    property LinesVertHorColor:TColor read _LinesVertHorColor write SetLinesVertHorColor default clGreen;
    property TimerEnabled:boolean read _TimerEnabled write SetTimerEnabled default false;

    property OnTmrTimer:TNotifyEvent read _OnTmrTimer write _OnTmrTimer;

    property Align;
    property Anchors;
    property Constraints;
    property Cursor;
    property DragCursor;
    property DragKind;
    property DragMode;
    property HelpContext;
    property Hint;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;

    property OnContextPopup;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
    property OnStartDock;

    property OnAlarmMin:KrAlarmMin read _AlarmMin write _AlarmMin;
    property OnAlarmMax:KrAlarmMax read _AlarmMax write _AlarmMax;
    property OnWarningMin:KrWarningMin read _WarningMin write _WarningMin;
    property OnWarningMax:KrWarningMax read _WarningMax write _WarningMax;
    property OnBorderMin:KrBorderMin read _BorderMin write _BorderMin;
    property OnBorderMax:KrBorderMax read _BorderMax write _BorderMax;
    property OnBeforePoint:KrBeforePoint read _BeforePoint write _BeforePoint;
    property OnAfterPoint:KrAfterPoint read _AfterPoint write _AfterPoint;
  end;

  KrAbout=class(TPropertyEditor)
  public
    procedure Edit; override;
    function GetAttributes: TPropertyAttributes; override;
    function GetValue: string; override;
  end;

procedure Register;

implementation
  constructor KrGraphic.Create(AOwner:TComponent);
  begin
    inherited Create(AOwner);

    BM:=TBitmap.Create;
    Tmr:=TTimer.Create(self);
    Tmr.Enabled:=false;
    Tmr.OnTimer:=TmrTimer;

    CONST_NOTDEFINED:=9223372036854775807;

    _Color:=clBlack;
    _Border:=boLine;
    _Interval:=1000;
    _LineColor:=clWhite;
    _ShadowColor:=clGray;
    _LightColor:=clWhite;
    _Lines:=liVertHor;
    _LinesVert:=10;
    _LinesHor:=10;
    _LinesVertHorColor:=clGreen;
    _TimerEnabled:=false;
  end;

  destructor KrGraphic.Destroy;
  var
    i:integer;
  begin
    BM.Free;
    Tmr.Free;

    for i:=1 to CONST_MAXELEMENTS do
      DeleteElement(i);

    inherited Destroy;
  end;

  procedure KrGraphic.Paint;
  begin
    BeginPaint;
  end;

  procedure KrGraphic.ResetPoints(Element:integer);
  var
    i:integer;
  begin
    for i:=1 to CONST_MAXPOINTS do
     Elements[Element]^.Points[i]:=CONST_NOTDEFINED;
  end;

  procedure KrGraphic.DeleteElement(Element:integer);
  begin
      if Elements[Element]<>Nil then Dispose(Elements[Element]);
  end;

  procedure KrGraphic.SetAlarmMin(Element:integer; Value:int64);
  begin
    if Elements[Element]=Nil then exit;
    Elements[Element]^.AlarmMin:=Value;
    BeginPaint;
  end;

  procedure KrGraphic.SetBorderMin(Element:integer; Value:int64);
  begin
    if Elements[Element]=Nil then exit;
    Elements[Element]^.BorderMin:=Value;
    BeginPaint;
  end;

  procedure KrGraphic.SetAlarmMax(Element:integer; Value:int64);
  begin
    if Elements[Element]=Nil then exit;
    Elements[Element]^.AlarmMax:=Value;
    BeginPaint;
  end;

  procedure KrGraphic.SetBorderMax(Element:integer; Value:int64);
  begin
    if Elements[Element]=Nil then exit;
    Elements[Element]^.BorderMax:=Value;
    BeginPaint;
  end;

  procedure KrGraphic.SetWarningMin(Element:integer; Value:int64);
  begin
    if Elements[Element]=Nil then exit;
    Elements[Element]^.WarningMin:=Value;
    BeginPaint;
  end;

  procedure KrGraphic.SetWarningMax(Element:integer; Value:int64);
  begin
    if Elements[Element]=Nil then exit;
    Elements[Element]^.WarningMax:=Value;
    BeginPaint;
  end;

  procedure KrGraphic.SetColorLine(Element:integer; Value:TColor);
  begin
    if Elements[Element]=Nil then exit;
    Elements[Element]^.Color:=Value;
    BeginPaint;
  end;

  procedure KrGraphic.PointValue(Element, Point:integer; Value:int64);
  begin
    if Elements[Element]=Nil then exit;
    if Assigned(_BeforePoint) then _BeforePoint(Self, Element, value);
    Elements[Element]^.Points[Point]:=Value;
    SendAlarms(Element, Value);
    if Assigned(_AfterPoint) then _AfterPoint(Self, Element, value);
    BeginPaint;
  end;

  function KrGraphic.AppendPointValue(Element:integer; value:int64):int64;
  var
    i:integer;
  begin
    Result:=CONST_NOTDEFINED;
    if Elements[Element]=Nil then exit;
    Result:=Elements[Element]^.Points[1];
    for i:=2 to CONST_MAXPOINTS do
      Elements[Element]^.Points[i-1]:=Elements[Element]^.Points[i];
    PointValue(Element, CONST_MAXPOINTS, Value);
  end;

  function KrGraphic.GetAlarmMin(Element:integer):int64;
  begin
    if Elements[Element]=nil then begin
      Result:=CONST_NOTDEFINED;
      exit;
    end;
    Result:=Elements[Element]^.AlarmMin;
  end;

  function KrGraphic.GetBorderMin(Element:integer):int64;
  begin
    if Elements[Element]=nil then begin
      Result:=CONST_NOTDEFINED;
      exit;
    end;
    Result:=Elements[Element]^.BorderMin;
  end;

  function KrGraphic.GetAlarmMax(Element:integer):int64;
  begin
    if Elements[Element]=nil then begin
      Result:=CONST_NOTDEFINED;
      exit;
    end;
    Result:=Elements[Element]^.AlarmMax;
  end;

  function KrGraphic.GetBorderMax(Element:integer):int64;
  begin
    if Elements[Element]=nil then begin
      Result:=CONST_NOTDEFINED;
      exit;
    end;
    Result:=Elements[Element]^.BorderMax;
  end;

  function KrGraphic.GetWarningMin(Element:integer):int64;
  begin
    if Elements[Element]=nil then begin
      Result:=CONST_NOTDEFINED;
      exit;
    end;
    Result:=Elements[Element]^.WarningMin;
  end;

  function KrGraphic.GetStatus(Element:integer):KrGStatus;
  begin
    if Elements[Element]=nil then begin
      Result:=stNone;
      exit;
    end;
    Result:=Elements[Element]^.Status;
  end;

  function KrGraphic.GetWarningMax(Element:integer):int64;
  begin
    if Elements[Element]=nil then begin
      Result:=CONST_NOTDEFINED;
      exit;
    end;
    Result:=Elements[Element]^.WarningMax;
  end;

  function KrGraphic.GetColorLine(Element:integer):TColor;
  begin
    if Elements[Element]=nil then begin
      Result:=CONST_NOTDEFINED;
      exit;
    end;
    Result:=Elements[Element]^.Color;
  end;

  function KrGraphic.PointValue(Element, Point:integer):int64;
  begin
    if Elements[Element]=nil then begin
      Result:=CONST_NOTDEFINED;
      exit;
    end;
    Result:=Elements[Element]^.Points[Point];
  end;

  procedure KrGraphic.SetTimerEnabled(value:boolean);
  begin
    _TimerEnabled:=value;
    Tmr.Enabled:=value;
  end;

  procedure KrGraphic.SetLinesVertHorColor(value:TColor);
  begin
    if _LinesVertHorColor=value then exit;
    _LinesVertHorColor:=value;
    BeginPaint;
  end;

  procedure KrGraphic.SetLinesHor(value:integer);
  begin
    if _LinesHor=value then exit;
    _LinesHor:=value;
    BeginPaint;
  end;

  procedure KrGraphic.SetLinesVert(value:integer);
  begin
    if _LinesVert=value then exit;
    _LinesVert:=value;
    BeginPaint;
  end;

  procedure KrGraphic.SetLines(value:KrGLines);
  begin
    if _Lines=value then exit;
    _Lines:=value;
    BeginPaint;
  end;

  procedure KrGraphic.SetLightColor(value:TColor);
  begin
    if _LightColor=value then exit;
    _LightColor:=value;
    BeginPaint;
  end;

  procedure KrGraphic.SetShadowColor(value:TColor);
  begin
    if _ShadowColor=value then exit;
    _ShadowColor:=value;
    BeginPaint;
  end;

  procedure KrGraphic.SetLineColor(value:TColor);
  begin
    if _LineColor=value then exit;
    _LineColor:=value;
    BeginPaint;
  end;

  procedure KrGraphic.SetColor(value:TColor);
  begin
    if _Color=value then exit;
    _Color:=value;
    BeginPaint;
  end;

  procedure KrGraphic.SetBorder(value:KrGBorder);
  begin
    if _Border=value then exit;
    _Border:=value;
    BeginPaint;
  end;

  procedure KrGraphic.SetInterval(value:integer);
  begin
    if value<>_Interval then begin
      _Interval:=value;
      Tmr.Interval:=_Interval;
    end;
  end;

  procedure KrGraphic.TmrTimer(Sender: TObject);
  begin
    if Assigned(_OnTmrTimer) then _OnTmrTimer(self);
  end;

  function KrGraphic.CreateElement(No:integer; BorderMin, BorderMax, AlarmMin, AlarmMax, WarningMin, WarningMax:int64; Color:TColor):boolean;
  begin
    result:=false;
    if (no<1) or (no>CONST_MAXELEMENTS) then exit;
    try
      if Elements[no]<>nil then Dispose(Elements[no]);
      New(Elements[no]);
      SetAlarmMin(no, AlarmMin);
      SetAlarmMax(no, AlarmMax);
      SetWarningMax(no, WarningMax);
      SetWarningMin(no, WarningMin);
      SetBorderMin(no, BorderMin);
      SetBorderMax(no, BorderMax);
      SetColorLine(no, Color);
      Elements[no]^.Status:=stNone;
      IsVisible(no, True);
      SetLength(Elements[no]^.Points, CONST_MAXPOINTS+1);
      ResetPoints(no);
    except
      exit;
    end;
    BeginPaint;
    result:=true;
  end;

  function KrGraphic.IsDefined(Element, Point:integer):boolean;
  begin
    result:=False;
    if Elements[Element]=nil then exit;
    Result:=not(Elements[Element]^.Points[point]=CONST_NOTDEFINED);
  end;

  procedure KrGraphic.IsVisible(Element:integer; Value:boolean);
  begin
    if Elements[Element]=nil then exit;
    Elements[Element]^.Visible:=value;
  end;

  function KrGraphic.IsVisible(Element:integer):boolean;
  begin
    Result:=False;
    if Elements[Element]=nil then exit;
    Result:=Elements[Element]^.Visible;
  end;

  procedure KrGraphic.PaintAlarmLines;
  var
    i:integer;
  begin
    for i:=1 to CONST_MAXELEMENTS do begin
      if IsVisible(i)=True then begin
        with BM.Canvas do begin
          if Elements[i]^.AlarmMin<>CONST_NOTDEFINED then begin
            Pen.Color:=GetColorLine(i);
            Pen.Style:=psDot;
            MoveTo(1, CalcPointX(i, GetAlarmMin(i)));
            LineTo(Width, CalcPointX(i, GetAlarmMin(i)));
          end;
          if Elements[i]^.AlarmMax<>CONST_NOTDEFINED then begin
            Pen.Color:=GetColorLine(i);
            Pen.Style:=psDot;
            MoveTo(1, CalcPointX(i, GetAlarmMax(i)));
            LineTo(Width, CalcPointX(i, GetAlarmMax(i)));
          end;
          if Elements[i]^.WarningMin<>CONST_NOTDEFINED then begin
            Pen.Color:=GetColorLine(i);
            Pen.Style:=psDashDotDot;
            MoveTo(1, CalcPointX(i, GetWarningMin(i)));
            LineTo(Width, CalcPointX(i, GetWarningMin(i)));
          end;
          if Elements[i]^.WarningMax<>CONST_NOTDEFINED then begin
            Pen.Color:=GetColorLine(i);
            Pen.Style:=psDashDotDot;
            MoveTo(1, CalcPointX(i, GetWarningMax(i)));
            LineTo(Width, CalcPointX(i, GetWarningMax(i)));
          end;
        end;
      end;
    end;
  end;

  procedure KrGraphic.PaintBorder;
  begin
    BM.Canvas.Pen.Style:=psSolid;
    case _Border of
      boDown: begin
          with BM.Canvas do begin
            Pen.Color:=_ShadowColor;
            MoveTo(0, Height);
            LineTo(0, 0);
            LineTo(Width-1, 0);
            Pen.Color:=_LightColor;
            LineTo(Width-1, Height-1);
            LineTo(0, Height-1);
          end;
        end;
      boUp:begin
          with BM.Canvas do begin
            Pen.Color:=_LightColor;
            MoveTo(0, Height);
            LineTo(0, 0);
            LineTo(Width-1, 0);
            Pen.Color:=_ShadowColor;
            LineTo(Width-1, Height-1);
            LineTo(0, Height-1);
          end;
        end;
      boLine: begin
          BM.Canvas.Brush.Color:=_LineColor;
          BM.Canvas.FrameRect(ClientRect);
        end;
    end;
  end;

  procedure KrGraphic.PaintPoints;
  var
    i, j, k:integer;
    l:boolean;
  begin
    BM.Canvas.Pen.Style:=psSolid;
    for i:=1 to CONST_MAXELEMENTS do begin
      if IsVisible(i)=True then begin
        with BM.Canvas do begin
          Pen.Color:=GetColorLine(i);
          l:=false;
          j:=0-integer(_Border<>boNone);
          for k:=CONST_MAXPOINTS-Width+2 to CONST_MAXPOINTS do begin
            inc(j);
            if IsDefined(i, k)=False then begin
              if l=true then
                LineTo(j, CalcPointX(i, PointValue(i, k)));
              l:=false;
            end else begin
              if l=false then begin
                MoveTo(j, CalcPointX(i, PointValue(i, k)));
                l:=true;
              end else
                LineTo(j, CalcPointX(i, PointValue(i, k)));
            end;
          end;
        end;
      end;
    end;
  end;

  procedure KrGraphic.PaintLines;
  var
    i, j:integer;
    d:double;
  begin
    if (_Lines=liVert) or (_Lines=liVertHor) then begin
      BM.Canvas.Pen.Color:=_LinesVertHorColor;
      BM.Canvas.Pen.Style:=psSolid;
      for i:=1 to _LinesVert do begin
        d:=Width;
        j:=Round(d/(_LinesVert+1)*i);
        BM.Canvas.MoveTo(j, 0+integer(_Border<>boNone));
        BM.Canvas.LineTo(j, Height-integer(_Border<>boNone));
      end;
    end;

    if (_Lines=liHor) or (_Lines=liVertHor) then begin
      BM.Canvas.Pen.Color:=_LinesVertHorColor;
      for i:=1 to _LinesHor do begin
        d:=Height;
        j:=Round(d/(_LinesHor+1)*i);
        BM.Canvas.MoveTo(0+integer(_Border<>boNone), j);
        BM.Canvas.LineTo(Width-integer(_Border<>boNone), j);
      end;
    end;
  end;

  function KrGraphic.CalcPointX(Element:integer; i:int64):integer;
  begin
    result:=round((i-GetBorderMin(Element))/(GetBorderMax(Element)-GetBorderMin(Element))*height);
  end;

  procedure KrGraphic.BeginPaint;
  begin
    BM.Width:=0;
    BM.Height:=0;
    BM.Width:=Width;
    BM.Height:=Height;
    BM.Canvas.Brush.Color:=_Color;
    BM.Canvas.FillRect(ClientRect);

    PaintBorder;
    PaintLines;
    PaintAlarmLines;
    PaintPoints;

    Canvas.CopyRect(ClientRect, BM.Canvas, ClientRect);
  end;

  procedure KrGraphic.SendAlarms(Element:integer; Value:int64);
  begin
    if Elements[Element]=nil then exit;

    Elements[Element]^.Status:=stNormal;
    if GetWarningMax(Element)<Value then begin
      if Assigned(_WarningMax) then _WarningMax(Self, Element, value);
      Elements[Element]^.Status:=stWarningMax;
    end;
    if GetAlarmMax(Element)<Value then begin
      if Assigned(_AlarmMax) then _AlarmMax(Self, Element, value);
      Elements[Element]^.Status:=stAlarmMax;
    end;
    if GetBorderMax(Element)<Value then begin
      if Assigned(_BorderMax) then _BorderMax(Self, Element, value);
      Elements[Element]^.Status:=stBorderMax;
    end;

    if GetWarningMin(Element)>Value then begin
      if Assigned(_WarningMin) then _WarningMin(Self, Element, value);
      Elements[Element]^.Status:=stWarningMin;
    end;
    if GetAlarmMin(Element)>Value then begin
      if Assigned(_AlarmMin) then _AlarmMin(Self, Element, value);
      Elements[Element]^.Status:=stAlarmMin;
    end;
    if GetBorderMin(Element)>Value then begin
      if Assigned(_BorderMin) then _BorderMin(Self, Element, value);
      Elements[Element]^.Status:=stBorderMin;
    end;
  end;

  procedure Register;
  begin
    RegisterComponents('kr...', [KrGraphic]);
    RegisterPropertyEditor(TypeInfo(KrAbout), KrGraphic, 'About', KrAbout);
  end;

  procedure KrAbout.Edit;
  begin
    MessageDlg('KrGraphic'+#13+#13+'Version 1.0, 2001'+#13+
               #13+
               'www.at-krays.com'+#13+
               'kray@at-krays.com',
                mtInformation, [mbOK], 0);
  end;

  function KrAbout.GetAttributes: TPropertyAttributes;
  begin
    Result := [paMultiSelect, paDialog, paReadOnly];
  end;

  function KrAbout.GetValue: string;
  begin
    Result := '(about)';
  end;
end.
